%!octave 
%
%
%       OS : GNU/Linux 4.2.5-1-ARCH 
% COMPILER : GNU Octave, version 4.0.0
%
%   AUTHOR : Evgeny S. Borisov
%
%  Glushkov Institute of Cybernetics
%  National Academy of Sciences of Ukraine
% 
%    http://www.mechanoid.kiev.ua
%  e-mail : nn@mechanoid.kiev.ua
% 
% $Id: aenc_trn_cdk.m,v 1.1 2016/03/19 19:29:19 mechanoid Exp $


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% обучение AENC методом  contrastive divergence
%
% L={Xl,Yl} - учебный набор
% С={Xc,Yc} - контрольный набор
% W - начальная матрица весов
% cfg - описание параметров сети сети
% trn_par - параметры обучения
%
function [W,er_hist] = aenc_trn_cdk(L,C,W,cfg,trn_par),
   ltp = 2 ; % номер ячейки cfg с типом ф-ции активации

   min_err = trn_par{1} ; % порог допустимой ошибки обучения
   max_iter = trn_par{2} ; % максимальное число эпох
   er_win_size = trn_par{3} ;  % размер окна контроля за ошибкой
   trsp = trn_par{4} ; % параметр скорости обучения
   moment_par = trn_par{5} ; % параметр момента 
   reg_par = trn_par{6} ; % параметр регуляризации
   batch_par = trn_par{7} ; % доля учебных примеров в одном пакете от общего количества примеров
   print_step = trn_par{8} ;% печатаем значение ошибки каждые print_step эпох 


   %er_win_size=5; % размер окна контроля за ошибкой
   %%trsp=9e-2; % скорость обучения (изменения весов)
   %trsp=8e-1; % скорость обучения (изменения весов)
   %reg_par=1e-2 ; % параметр регуляризации
   %moment_par=6e-1; % параметр момента 
   %batch_par=0.7; % доля учебных примеров в одном пакете от общего количества примеров

   X=L{1} ; % учебный набор

   num_examples=size(X,2) ; % количество объектов в учебном наборе
   batch_size=round(num_examples*batch_par); % количество учебных примеров в одном пакете 


   if (length(cfg)!=3 ) || (cfg{1}{1}!=cfg{3}{1}),
	  printf("[e] неправильная конфигурация автоэнкодера \n");
	  exit -1
   endif

   tname="contrastive divergence";
   printf("[i] обучение методом %s\n",tname);
   save("-text","result/tname.txt", "tname" ) ;

 
   er_hist = [er_win_size:-1:1]*100.0 ; % история значений ошибки обучения

   dW = mlp_weight_set(cfg,0.0) ; % начальные значения изменений весов

   for iter=1:max_iter,

	  er_hist(end+1) = mlp_loss(C,W,cfg,reg_par); % вычисляем ошибку сети на контрольном наборе

	  if mod(iter,print_step)<1, % печатаем значение ошибки N эпох 
		 printf("[i]   %i/%i : %f / %f   , %f     \r",iter,max_iter,er_hist(end),min_err,trsp ) ; 
      endif
 	 
	  if trn_stop_condition(er_hist,min_err,er_win_size), break ; endif % условия остановки обучения
 
	  % случайный пакет из учебных примеров
      batch_idx = randperm(num_examples)(1:batch_size); 
	  Xb = X(:,batch_idx) ;

	  g = grad(Xb,W,cfg); % значение градиента

	  dW = weight_change(g,dW,W,trsp,reg_par,moment_par); % изменение весов 
	  W = weight_correct(W,dW);

	  trsp = max(trsp*0.99,3e-2);
	 
   endfor

   printf("[i]   %i/%i : %f / %f   , %f     \n",iter,max_iter,er_hist(end),min_err,trsp ) ; 

   er_hist(1:er_win_size)=[]; % удаляем начальные значения ошибки из истории 

   if iter>(max_iter-1) , printf("[w] достигнут порог количества циклов обучения         \n") ; endif

endfunction


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% градиент функции энергии
%
function retval = grad(X,W,cfg),
	  
   [V,H] = sampling(X,W,cfg); % вычислить мат.ожидание слоёв (sampling)

   gwn = ( V{1}*(H{1})' - V{2}*(H{2})' )' ;
   gbv = sum( H{1} - H{2} , 2 ) ;
   gbh = sum( V{1} - V{2} , 2 ) ;


   ng = norm( [ gwn(:) ; gbv ; gbh ]);
   retval = { gwn/ng, gbv/ng, gbh/ng } ;

   % ex_numb = size(V(:,:,1),2);
   % retval = { gwn, gbv/ex_numb, gbh/ex_numb } ;
   %retval={ gwn, gbv, gbh } ;
endfunction

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function  [V,H ] = sampling(X,W,cfg),
   n = 2 ; % количество циклов sampling
   ltp=2; % номер ячейки cfg с типом ф-ции активации


   V={};H={};

   V{1} = X ; % сохраняем начальное состояние первого слоя
   l=mlp_layers(X, W, cfg );% вычисляем состояние скрытого слоя
   H{1} = mlp_activate( l{2}, cfg{2}{ltp} ) ; % сохраняем выход скрытого слоя

   % выполняем n циклов сети
   for i=1:n, 
	  X = mlp_activate( l{end}, cfg{end}{ltp} ) ; % перенаправляем выход на вход
	  l=mlp_layers(X, W, cfg ); % вычисляем состояния слоёв
   endfor

   V{2} = X ; % сохраняем конечное состояние первого слоя
   H{2} = mlp_activate( l{2}, cfg{2}{ltp} ) ; % сохраняем выход скрытого слоя

endfunction


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% вычисляем изменение весов
% 
% g - градиент функции ошибки
% dW - пред.изменение весов
% W - весовая матрица
% trsp - скорость обучения
% rpar - параметр регуляризации
% mpar - параметр момента
%
function retval = weight_change(g,dW,W,trsp,rpar,mpar),
   % коррекция весовых матриц слоёв с поправками момента и регуляризации
   dW{1}{1} = trsp*( g{1} + mpar*dW{1}{1} + rpar*W{1}{1} );

   dW{1}{2} = trsp*(g{2} + mpar*dW{1}{2} ) ;
   dW{2}{2} = trsp*(g{3} + mpar*dW{2}{2} ) ;

   retval=dW;
endfunction


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval = weight_correct(W,dW),
   % коррекция весовых матриц слоёв с поправками момента и регуляризации
   W{1}{1} += dW{1}{1};
   W{2}{1} = (W{1}{1})';

   W{1}{2} += dW{1}{2};
   W{2}{2} += dW{2}{2};

   retval=W;
endfunction





