%!octave 
%
%       OS : GNU/Linux 4.2.5-1-ARCH 
% COMPILER : GNU Octave, version 4.0.0
%
%   AUTHOR : Evgeny S. Borisov
%
%  Glushkov Institute of Cybernetics
%  National Academy of Sciences of Ukraine
% 
%    http://www.mechanoid.kiev.ua
%  e-mail : nn@mechanoid.kiev.ua
% 
% $Id: mlp_trn_cg.m,v 1.1 2016/03/19 19:02:54 mechanoid Exp $

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% обучение ИНС методом сопряжённого градиента 
%
% L={Xl,Yl} - учебный набор
% С={Xc,Yc} - контрольный набор
% W - начальная матрица весов
% cfg - описание параметров сети сети
% trn_par - параметры обучения
%
function [W,er_hist]=mlp_trn_cg(L,C,W,cfg,trn_par),
   ltp = 2 ; % номер ячейки cfg с типом ф-ции активации

   min_err = trn_par{1} ; % порог допустимой ошибки обучения
   max_iter = trn_par{2} ; % максимальное число эпох
   er_win_size = trn_par{3} ;  % размер окна контроля за ошибкой
   trsp0 = trn_par{4} ; % параметр скорости обучения
   moment_par = trn_par{5} ; % параметр момента 
   reg_par = trn_par{6} ; % параметр регуляризации
   print_step = trn_par{8} ; % печатаем значение ошибки print_step эпох

   %er_win_size=5; % размер окна контроля за ошибкой
   %trsp0=1e-2; % скорость обучения (изменения весов)
   %reg_par = 1e-4 ; % параметр регуляризации
   %moment_par=9e-2; % параметр момента 
   
   % общеее количество весов сети
   par_num=0; for i=1:length(cfg)-1, par_num+=(cfg{i}{1}+1)*cfg{i+1}{1}; endfor
   
   mcs=round(par_num*0.25); % максимальное количество шагов с сопряжением

   tname="сопряжённые градиенты";
   printf("[i] обучение методом %s\n",tname);
   save("-text","result/tname.txt", "tname" ) ;

 
   er_hist = [er_win_size:-1:1]*100.0 ; % история значений ошибки обучения

   dW = mlp_weight_set(cfg,0.0) ; % начальные (нулевые) значения изменения весов

   p = g = mlp_backprop_grad(L,W,cfg); % значение градиента и начальное направление изменения параметров

   cs=1; % количество шагов с сопряжением

   for iter=1:max_iter,
      
	  % выбираем оптимальный шаг обучения для текущего градиента
	  trsp = fminbnd(@(a)err_value(a,p,W,dW,C,cfg,reg_par,moment_par), trsp0, 1.0 ); 

	  dW = weight_change(p,dW,W,trsp,reg_par,moment_par);% корректируем веса 
	  W = mlp_weight_add(W,dW,-1.0);% корректируем веса 

	  er_hist(end+1) = mlp_loss(C,W,cfg,reg_par); % вычисляем ошибку сети на контрольном наборе

	  if mod(iter,print_step)<1, % печатаем значение ошибки print_step эпох 
		 printf("[i]   %i/%i : %f / %f   , %f     \r",iter,max_iter,er_hist(end),min_err,trsp ) ; 
      endif

 	  if trn_stop_condition(er_hist,min_err,er_win_size), break ; endif % условия остановки обучения
 
	  g0 = g; % сохраняем старое значение градиента
	  g = mlp_backprop_grad(L,W,cfg); % значение градиента

	  if(cs<mcs),% считаем коэффициент сопряжения 
		 cs+=1; 
		 vg=mlp_weight_unpack(g) ; 
		 vg0=mlp_weight_unpack(g0);
		 b=(vg'*vg)/(vg0'*vg0); % метод Флетчера-Ривса
		 %b=( vg'*(vg-vg0) ) ./ (vg0'*vg0) ; % метод Полака-Рибьера
		 %vp=mlp_weight_unpack(p); b=( vg'*(vg-vg0) ) ./ (-vp'*vg0) ; 
	  else
		 % метод предусматривает периодический сброс, 
		 % т.е. использование итерации наискорейшего спуска через каждые mcs шагов
		 b=0.0; cs=1; % сбрасываем сопряженный градиент 
	  endif

	  p = conj_dir(g,p,b); % новое сопряжённое направление  p=-g+b*p
   endfor

   printf("[i]   %i/%i : %f / %f   , %f     \n",iter,max_iter,er_hist(end),min_err,trsp ) ; 

   er_hist(1:er_win_size)=[]; % удаляем начальные значения ошибки из истории 

   if iter>(max_iter-1) , printf("[w] достигнут порог количества циклов обучения         \n") ; endif

endfunction


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval=err_value(a,p,W,dW,C,cfg,reg_par,moment_par),
   dW = weight_change(p,dW,W,a,reg_par,moment_par);
   W = mlp_weight_add(W,dW,-1.0); % корректируем веса 
   er = mlp_loss(C,W,cfg,reg_par); % вычисляем ошибку сети на контрольном наборе
   retval = er ;
endfunction


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% вычисляем изменение весов
% 
% g - градиент функции ошибки
% dW - пред.изменение весов
% W - весовая матрица
% trsp - скорость обучения
% rpar - параметр регуляризации
% mpar - параметр момента
%
function retval = weight_change(g,dW,W,trsp,rpar,mpar),
   s=length(W);
   for i=1:s,
	  % коррекция весовых матриц слоёв с поправками момента и регуляризации
      dW{i}{1} = trsp*( g{i}{1} + rpar*W{i}{1} )+ mpar*dW{i}{1} ;
      dW{i}{2} = trsp*g{i}{2} + mpar*dW{i}{2} ;
   endfor
   retval=dW;
endfunction


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% новое сопряжённое направление  p=-g+b*p;
function retval=conj_dir(g,p,b), 
   r={};
   for i=1:length(g),
	  for j=1:2,
		 r{i}{j} = g{i}{j} + p{i}{j}.*b;
	  endfor
   endfor
   retval=r;
endfunction


