%!octave 
%
%       OS : GNU/Linux 4.2.5
% COMPILER : GNU Octave, version 4.0.0
%
%   AUTHOR : Evgeny S. Borisov
%
%  Glushkov Institute of Cybernetics
%  National Academy of Sciences of Ukraine
% 
%    http://www.mechanoid.kiev.ua
%  e-mail : nn@mechanoid.kiev.ua
% 
% $Id: main.m,v 1.3 2016/03/20 10:17:44 mechanoid Exp mechanoid $

addpath("lib/data-proc");
addpath("lib/mlp");
addpath("lib/aenc");

split_long_rows(0) ; % set nowrap

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

printf("[i] загружаем данные\n"); % L учебная выборка; C - контрольная выборка ; T - тестовая выборка

[ fl,fc,ft ] = data_path_img();

L=data_load_img(fl); % L={ X,Y,img_size,files_list}
C=data_load_img(fc); % L={ X,Y,img_size,files_list}

%L{1} = (L{1} > 0.5 ); C{1} = (C{1} > 0.5 );


printf("[i] загружено %i: учебных: %i контрольных: %i \n",size(L{2},2)+size(C{2},2),size(L{2},2),size(C{2},2));

sz_in=size(L{1},1);
sz_out=size(L{2},1);

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% { размер слоя, тип активации }  , тип активации: input tanh expsig relu softmax
cfg = { 
   { sz_in,"input" }, 
   { 100,"relu" }, 
   { 100,"relu" }, 
   { 100,"relu" }, 
   { 100,"relu" }, 
   { 10,"relu" }, 
   { sz_out,"softmax" } 
 };
 
mlp_print_cfg(cfg);

save("-text","result/cfg.txt", "cfg" ) ;

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% предобучение MLP

W = mlp_weight_init(cfg) ; 

ptrn_par={
   1e-1, % порог допустимой ошибки обучения
   1500, % максимальное число эпох
   5,  % размер окна контроля за ошибкой
   5e-1, % параметр скорости обучения
   5e-2, % параметр момента 
   1e-2, % параметр регуляризации
   0.7, % доля учебных примеров в одном пакете (mini-batch) от общего количества примеров
   10  % печатаем значение ошибки каждые print_step эпох 
};
save("-text","result/ptrn_par.txt", "ptrn_par" ) ;


W = mlp_pretrain(L,W,cfg,ptrn_par); % предобучение

save("-text","result/w0.txt", "W" ) ;
%load("data1/w.txt")

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% обучение MLP

printf("[i] обучение ...\n"); % load("data1/w.txt");

trn_par={
   1e-1, % порог допустимой ошибки обучения
   200, % максимальное число эпох
   3,  % размер окна контроля за ошибкой
   3e-2, % параметр скорости обучения
   5e-3, % параметр момента 
   1e-4, % параметр регуляризации
   0.7, % доля учебных примеров в одном пакете (mini-batch) от общего количества примеров
   10  % печатаем значение ошибки каждые print_step эпох 
};
save("-text","result/trn_par.txt", "trn_par" ) ;


[W,er]=mlp_trn_gd(L,C,W,cfg,trn_par);
%[W,er]=mlp_trn_cg(L,C,W,cfg,trn_par);
%[W,er]=mlp_trn_qprop(L,C,W,cfg,trn_par);
%[W,er]=mlp_trn_rprop(L,C,W,cfg,trn_par);


save("-text","result/w.txt", "W" ) ;
save("-text","result/er.txt", "er" ) ;
 
plot_error(er,0); % рисуем историю изменения ошибки обучения

weight_visualizer(W{1}{1},L{3}); % рисуем матрицу весов

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
 
printf("[i] тест                   \n");

T=data_load_img(ft); % L={ X,Y,img_size,files_list}
printf("[i] загружено тестовых: %i \n",size(T{2},2));

[ er,o ] = mlp_loss(T,W,cfg,0.0) ;
%save("-text","result/o.txt", "o" ) ;

er

[ p,n ]=max(o); % ответ (n) и вероятность принадлежности к классу (p)

n-=1 ;

save("-text","result/answ-n.txt", "n" ) ;
save("-text","result/answ-p.txt", "p" ) ;

[ v,vi ]=max(T{2}); % правильный ответ (vi)
v=(n!=(vi-1)); % ошибки

r=[ sum(v), length(v)];

printf("[i] ошибок на тесте %i из %i\n", r(1), r(2));

save("-text","result/answ-e.txt", "r" ) ;
save("-text","result/answ-ev.txt", "v" ) ;




% предобучение:
%    определить размеры слоёв
%    повторить для всех соседних пар слоёв  
% 	  обучить AEN
% 	  сохранить веса
% 	  генерировать следующий набор данных (выходы скрытого слоя)
% 
% собрать mlp из получившихся матриц весов 
% обучить mlp
% 
% тестировать mlp


